from google.cloud.bigquery.table import RowIterator
from tabulate import tabulate


def print_results(it: RowIterator, format_args: dict = None):
    if format_args is None:
        format_args = {
            "tablefmt": "pretty",
            "missingval": "null",
            "stralign": 'right',
        }
    headers = {field.name: field.name for field in it.schema}
    rows = list(it)
    print(tabulate(rows, headers=headers, **format_args))




from typing import Union

from bigquery_frame.dataframe import cols_to_str, Column
from bigquery_frame.functions import __str_to_col


def when(condition: Column, value: Column) -> Column:
    """Evaluates a list of conditions and returns one of multiple possible result expressions.
    If :func:`Column.otherwise` is not invoked, None is returned for unmatched conditions.

    :param condition: a boolean :class:`Column` expression.
    :param value: a :class:`Column` expression.
    :return:
    """
    # TODO


def hash(*cols: Union[str, Column]) -> Column:
    cols = __str_to_col(cols)
    return Column(f"FARM_FINGERPRINT(TO_JSON_STRING(STRUCT({cols_to_str(cols)})))")


def sort_array(col: Union[str, Column]) -> Column:
    col = __str_to_col(col)
    return Column(f"ARRAY(SELECT x FROM UNNEST({col}) AS x ORDER BY x)")



from bigquery_frame import DataFrame, BigQueryBuilder
from bigquery_frame.auth import get_bq_client


def sort_columns(df: DataFrame) -> DataFrame:
    """Returns a new DataFrame where the order of columns has been sorted

    Examples:
    >>> bq = BigQueryBuilder(get_bq_client())
    >>> df = bq.sql('''SELECT 1 as b, 1 as a, 1 as c''')
    >>> df.printSchema()
    root
     |-- b: INTEGER (NULLABLE)
     |-- a: INTEGER (NULLABLE)
     |-- c: INTEGER (NULLABLE)
    <BLANKLINE>
    >>> sort_columns(df).printSchema()
    root
     |-- a: INTEGER (NULLABLE)
     |-- b: INTEGER (NULLABLE)
     |-- c: INTEGER (NULLABLE)
    <BLANKLINE>
    """
    return df.select(*sorted(df.columns))




from bigquery_frame.transformations_impl.pivot_unpivot import pivot, unpivot
from bigquery_frame.transformations_impl.flatten import flatten, flatten_schema
from bigquery_frame.transformations_impl.sort_columns import sort_columns
from bigquery_frame.transformations_impl.union_dataframes import union_dataframes
from bigquery_frame.transformations_impl.analyze import analyze




# This file may be edited by the user.
# Please read AUTH.md first before changing anything.

# Method 1. Set this variable here or set it as an environment variable
GCP_PROJECT = "Name of your BigQuery project"


# Method 2. Set this variable here or set it as an environment variable
GCP_CREDENTIALS_PATH = "Path to your service account credentials json file"




from typing import List

from bigquery_frame import DataFrame
from bigquery_frame.dataframe import quote


def union_dataframes(dfs: List[DataFrame]) -> DataFrame:
    """Returns the union between multiple DataFrames"""
    if len(dfs) == 0:
        raise ValueError("input list is empty")
    query = "\nUNION ALL\n".join([f"  SELECT * FROM {quote(df._alias)}" for df in dfs])
    return DataFrame(query, alias=None, bigquery=dfs[0].bigquery, deps=dfs)




from bigquery_frame import BigQueryBuilder
from bigquery_frame.auth import get_bq_client
from bigquery_frame.column import Column as StrictColumn
from bigquery_frame.dataframe import cols_to_str, DataFrame
from typing import Union, List, Iterable

Column = Union[str, StrictColumn]


def __str_to_col(args: Union[Iterable[Column], Column]) -> Union[List[StrictColumn], StrictColumn]:
    """Converts string arguments to Column types"""
    if isinstance(args, str):
        return StrictColumn(expr=args)
    elif isinstance(args, list):
        return [__str_to_col(arg) for arg in args]
    else:
        return args


def col(expr: str) -> StrictColumn:
    return StrictColumn(expr=expr)


def lit(val: object) -> StrictColumn:
    if val is None:
        return StrictColumn("NULL")
    if type(val) == str:
        return StrictColumn(f"'{val}'")
    if type(val) in [bool, int, float]:
        return StrictColumn(str(val))
    raise ValueError(f'lit({val}): The type {type(val)} is not supported yet.')


def struct(*cols: Column) -> Column:
    return StrictColumn(f"STRUCT({cols_to_str(cols)})")


def isnull(col: Column) -> Column:
    return StrictColumn(f"{col} IS NULL")


def count(col: Union[Column, str]):
    """Aggregate function: returns the number of rows where the specified column is not null

    >>> df = __get_test_df_1()
    >>> df.show()
    +------+------+
    | col1 | col2 |
    +------+------+
    |    1 |    a |
    |    1 |    b |
    |    2 | null |
    +------+------+
    >>> from bigquery_frame import functions as f
    >>> df.select(
    ...   f.count(f.lit(1)).alias('count_1'),
    ...   f.count('col1').alias('count_col1'),
    ...   f.count('col2').alias('count_col2'),
    ...   f.count('*').alias('count_star')
    ... ).show()
    +---------+------------+------------+------------+
    | count_1 | count_col1 | count_col2 | count_star |
    +---------+------------+------------+------------+
    |       3 |          3 |          2 |          3 |
    +---------+------------+------------+------------+

    """
    col = __str_to_col(col)
    return StrictColumn(f"COUNT({col.expr})")


def count_distinct(col: Union[Column, str]):
    """Aggregate function: returns the number of distinct non-null values

    >>> df = __get_test_df_1()
    >>> df.show()
    +------+------+
    | col1 | col2 |
    +------+------+
    |    1 |    a |
    |    1 |    b |
    |    2 | null |
    +------+------+
    >>> from bigquery_frame import functions as f
    >>> df.select(
    ...   f.count_distinct('col1').alias('count_distinct_col1'),
    ...   f.count_distinct('col2').alias('count_distinct_col2'),
    ... ).show()
    +---------------------+---------------------+
    | count_distinct_col1 | count_distinct_col2 |
    +---------------------+---------------------+
    |                   2 |                   2 |
    +---------------------+---------------------+

    """
    col = __str_to_col(col)
    return StrictColumn(f"COUNT(DISTINCT {col.expr})")


def min(col: Union[Column, str]):
    """Aggregate function: returns the minimum value of the expression in a group.

    >>> df = __get_test_df_1()
    >>> df.show()
    +------+------+
    | col1 | col2 |
    +------+------+
    |    1 |    a |
    |    1 |    b |
    |    2 | null |
    +------+------+
    >>> from bigquery_frame import functions as f
    >>> df.select(
    ...   f.min('col1').alias('min_col1'),
    ...   f.min('col2').alias('min_col2'),
    ... ).show()
    +----------+----------+
    | min_col1 | min_col2 |
    +----------+----------+
    |        1 |        a |
    +----------+----------+

    """
    col = __str_to_col(col)
    return StrictColumn(f"MIN({col.expr})")


def max(col: Union[Column, str]):
    """Aggregate function: returns the maximum value of the expression in a group.

    >>> df = __get_test_df_1()
    >>> df.show()
    +------+------+
    | col1 | col2 |
    +------+------+
    |    1 |    a |
    |    1 |    b |
    |    2 | null |
    +------+------+
    >>> from bigquery_frame import functions as f
    >>> df.select(
    ...   f.max('col1').alias('max_col1'),
    ...   f.max('col2').alias('max_col2'),
    ... ).show()
    +----------+----------+
    | max_col1 | max_col2 |
    +----------+----------+
    |        2 |        b |
    +----------+----------+

    """
    col = __str_to_col(col)
    return StrictColumn(f"MAX({col.expr})")


def expr(expr: str):
    """Parses the expression string into the column that it represents.

    >>> df = __get_test_df_1()
    >>> df.show()
    +------+------+
    | col1 | col2 |
    +------+------+
    |    1 |    a |
    |    1 |    b |
    |    2 | null |
    +------+------+
    >>> from bigquery_frame import functions as f
    >>> df.select(
    ...   f.expr('COALESCE(col2, CAST(col1 as STRING)) as new_col')
    ... ).show()
    +---------+
    | new_col |
    +---------+
    |       a |
    |       b |
    |       2 |
    +---------+

    """
    return StrictColumn(expr)


def __get_test_df_1() -> DataFrame:
    bq = BigQueryBuilder(get_bq_client())
    query = """
        SELECT * FROM UNNEST ([
            STRUCT(1 as col1, "a" as col2),
            STRUCT(1 as col1, "b" as col2),
            STRUCT(2 as col1, NULL as col2)
       ])
    """
    return bq.sql(query)




from bigquery_frame import DataFrame


def export_analyze(df: DataFrame):
    # TODO: finish this
    df.withColumn("approx_top_100", f.expr('''ARRAY(
      SELECT CONCAT('(', SUBSTRING(x.value, 0, 10), CASE WHEN LENGTH(x.value) > 10 THEN '...' ELSE '' END, ',', CAST(x.count as STRING), ')')
      FROM UNNEST(approx_top_100) x
    )'''), replace=True)
    df = df.withColumn("approx_top_100", f.expr('''CONCAT('[', ARRAY_TO_STRING(approx_top_100, ", ") , ']')'''), replace=True)
    df.show()





from typing import List

from google.cloud.bigquery import SchemaField

from bigquery_frame import DataFrame, BigQueryBuilder
from bigquery_frame.auth import get_bq_client
from bigquery_frame import functions as f
from bigquery_frame.dataframe import strip_margin, quote
from bigquery_frame.transformations_impl.flatten import flatten_schema
from bigquery_frame.transformations_impl.union_dataframes import union_dataframes


def _unnest_column(df: DataFrame, col: str):
    """Recursively unnest a :class:`DataFrame`'s column

    >>> bq = BigQueryBuilder(get_bq_client())
    >>> df = bq.sql('SELECT 1 as id, [STRUCT(2 as a, [STRUCT(3 as c, 4 as d)] as b, [5, 6] as e)] as s1, STRUCT(7 as f) as s2')
    >>> [col.name for col in flatten_schema(df.schema, explode=True)]
    ['id', 's1!.a', 's1!.b!.c', 's1!.b!.d', 's1!.e!', 's2.f']
    >>> _unnest_column(df, 'id').show()
    +----+--------------------------------------------------+----------+
    | id |                                               s1 |       s2 |
    +----+--------------------------------------------------+----------+
    |  1 | [{'a': 2, 'b': [{'c': 3, 'd': 4}], 'e': [5, 6]}] | {'f': 7} |
    +----+--------------------------------------------------+----------+
    >>> _unnest_column(df, 's1!.a').show()
    +---+
    | a |
    +---+
    | 2 |
    +---+
    >>> _unnest_column(df, 's1!.b!.c').show()
    +---+
    | c |
    +---+
    | 3 |
    +---+
    >>> _unnest_column(df, 's1!.b!.d').show()
    +---+
    | d |
    +---+
    | 4 |
    +---+
    >>> _unnest_column(df, 's1!.e!').show()
    +---+
    | e |
    +---+
    | 5 |
    | 6 |
    +---+
    >>> _unnest_column(df, 's2.f').show()
    +----+--------------------------------------------------+----------+
    | id |                                               s1 |       s2 |
    +----+--------------------------------------------------+----------+
    |  1 | [{'a': 2, 'b': [{'c': 3, 'd': 4}], 'e': [5, 6]}] | {'f': 7} |
    +----+--------------------------------------------------+----------+

    :param df:
    :param col:
    :return:
    """
    def build_cross_join_statement(split: List[str]):
        previous = ""
        counter = 1
        for sub in split[:-1]:
            if sub[0] == ".":
                struct = previous + sub
            else:
                struct = sub
            alias = f"_unnest_{counter}"
            previous = alias
            counter += 1
            yield f"CROSS JOIN UNNEST({struct}) as {alias}"

    if "!" in col:
        split = col.split("!")
        cross_joins = list(build_cross_join_statement(split))
        cross_join_str = "\n".join(cross_joins)
        last = split[-1]
        col = f"_unnest_{len(cross_joins)}"
        if "." in last:
            col += last
        else:
            col += " as " + split[-2].replace(".", "")

        query = strip_margin(f"""
            |SELECT 
            |  {col}
            |FROM {quote(df._alias)}
            |{cross_join_str}""")
        return df._apply_query(query)
    else:
        return df


def _analyze_column(df: DataFrame, schema_field: SchemaField, col_num: int):
    col = schema_field.name
    if "!" in col:
        if col[-1] == "!":
            col = col.split("!")[-2]
        else:
            col = col.split(".")[-1]
    df = _unnest_column(df, schema_field.name)
    res = df.select(
        f.lit(col_num).alias("column_number"),
        f.lit(schema_field.name).alias("column_name"),
        f.lit(schema_field.field_type).alias("column_type"),
        f.count(f.lit(1)).alias("count"),
        f.count_distinct(col).alias("count_distinct"),
        (f.count(f.lit(1)) - f.count(col)).alias("count_null"),
        f.min(col).asType("STRING").alias("min"),
        f.max(col).asType("STRING").alias("max"),
        f.expr(f"APPROX_TOP_COUNT(COALESCE(CAST({col} as STRING), 'NULL'), 100)").alias("approx_top_100")
    )
    return res


def __chunks(lst: List, n: int):
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(lst), n):
        yield lst[i:i + n]


def analyze(df: DataFrame, _chunk_size: int = 50):
    """Analyze a DataFrame by computing various stats for each column.

    It returns a DataFrame with one row per column and the following columns:

    - column_name: Name of the column
    - column_type: Type of the column
    - count: Number of rows in the column, it is equal to the number of rows in the table, except for columns nested
      inside arrays for which it may be different
    - count_distinct: Number of distinct values
    - count_null: Number of null values
    - min: smallest value
    - max: largest value
    - approx_top_100: Top 100 most frequent values with their respective count

    Implementation details:
    -----------------------

    - Structs are flattened with a `.` in their name.
    - Arrays are unnested with a `!` character in their name, which is why they may have a different count.
    - Null values are not counted in the count_distinct column.

    Troubleshooting:
    -----------------------
    When working on entreprise-level data, BigQuery might return a 400 error with the following message.
    "Not enough resources for query planning - too many subqueries or query is too complex"

    This may happen in two cases:
    - If the input DataFrame is already the result of a complex calculation, it is recommended to save it to a
      permanent or temporary table first. You can save it to a temporary table by using `df = df.persist()`
    - When analyzing DataFrames with hundreds of columns, the query generated by analyze() might get too complex
      for BigQuery. To avoid this, we split the query into smaller parts that are persisted.
      The optional argument `_chunk_size` (default = 50) determines the number of columns analyzed per chunk.
      If you already persisted the input DataFrame and the same error persist, you can try lowering the value
      of `_chunk_size`, but beware that the lower this value, the slower the analyzis will be.

    >>> df = __get_test_df()
    >>> df.show()
    +----+------------+---------------------+--------------------------------------------+
    | id |       name |               types |                                  evolution |
    +----+------------+---------------------+--------------------------------------------+
    |  1 |  Bulbasaur | ['Grass', 'Poison'] | {'can_evolve': True, 'evolves_from': None} |
    |  2 |    Ivysaur | ['Grass', 'Poison'] |    {'can_evolve': True, 'evolves_from': 1} |
    |  3 |   Venusaur | ['Grass', 'Poison'] |   {'can_evolve': False, 'evolves_from': 2} |
    |  4 | Charmander |            ['Fire'] | {'can_evolve': True, 'evolves_from': None} |
    |  5 | Charmeleon |            ['Fire'] |    {'can_evolve': True, 'evolves_from': 4} |
    |  6 |  Charizard |  ['Fire', 'Flying'] |   {'can_evolve': False, 'evolves_from': 5} |
    |  7 |   Squirtle |           ['Water'] | {'can_evolve': True, 'evolves_from': None} |
    |  8 |  Wartortle |           ['Water'] |    {'can_evolve': True, 'evolves_from': 7} |
    |  9 |  Blastoise |           ['Water'] |   {'can_evolve': False, 'evolves_from': 8} |
    +----+------------+---------------------+--------------------------------------------+
    >>> df = analyze(df)
    Analyzing 5 columns ...
    >>> df.withColumn("approx_top_100", f.expr("approx_top_100[OFFSET(0)]"), replace=True).show()
    +------------------------+-------------+-------+----------------+------------+-----------+-----------+------------------------------------+
    |            column_name | column_type | count | count_distinct | count_null |       min |       max |                     approx_top_100 |
    +------------------------+-------------+-------+----------------+------------+-----------+-----------+------------------------------------+
    |                     id |     INTEGER |     9 |              9 |          0 |         1 |         9 |         {'value': '1', 'count': 1} |
    |                   name |      STRING |     9 |              9 |          0 | Blastoise | Wartortle | {'value': 'Bulbasaur', 'count': 1} |
    |                 types! |      STRING |    13 |              5 |          0 |      Fire |     Water |     {'value': 'Grass', 'count': 3} |
    |   evolution.can_evolve |     BOOLEAN |     9 |              2 |          0 |     false |      true |      {'value': 'true', 'count': 6} |
    | evolution.evolves_from |     INTEGER |     9 |              6 |          3 |         1 |         8 |      {'value': 'NULL', 'count': 3} |
    +------------------------+-------------+-------+----------------+------------+-----------+-----------+------------------------------------+

    :param df:
    :param _chunk_size: (default: 50) Try reducing this number if you get this error from BigQuery: "Not enough resources for query planning"
    :return:
    """
    flat_schema = flatten_schema(df.schema, explode=True)
    col_dfs = [_analyze_column(df, schema_field, num) for num, schema_field in enumerate(flat_schema)]

    nb_cols = len(col_dfs)
    print(f"Analyzing {nb_cols} columns ...")
    if nb_cols <= _chunk_size:
        union_df = union_dataframes(col_dfs)
    else:
        from tqdm import tqdm
        big_dfs = [
            union_dataframes(chunk).persist()
            for chunk in tqdm(list(__chunks(col_dfs, _chunk_size)))
        ]
        union_df = union_dataframes(big_dfs)
    # For some reason, `union_df.orderBy("column_number").drop("column_number")`
    # does not preserve the ordering, but this does:
    query = f"""SELECT * EXCEPT(column_number) FROM {quote(union_df._alias)} ORDER BY column_number"""
    return union_df._apply_query(query)


def __get_test_df() -> DataFrame:
    bq = BigQueryBuilder(get_bq_client())
    query = """
        SELECT 
            *
        FROM UNNEST ([
            STRUCT(
                1 as id, "Bulbasaur" as name, ["Grass", "Poison"] as types, 
                STRUCT(TRUE as can_evolve, NULL as evolves_from) as evolution
            ),
            STRUCT(
                2 as id, "Ivysaur" as name, ["Grass", "Poison"] as types, 
                STRUCT(TRUE as can_evolve, 1 as evolves_from) as evolution
            ),
            STRUCT(
                3 as id, "Venusaur" as name, ["Grass", "Poison"] as types, 
                STRUCT(FALSE as can_evolve, 2 as evolves_from) as evolution
            ),
            STRUCT(
                4 as id, "Charmander" as name, ["Fire"] as types, 
                STRUCT(TRUE as can_evolve, NULL as evolves_from) as evolution
            ),
            STRUCT(
                5 as id, "Charmeleon" as name, ["Fire"] as types, 
                STRUCT(TRUE as can_evolve, 4 as evolves_from) as evolution
            ),
            STRUCT(
                6 as id, "Charizard" as name, ["Fire", "Flying"] as types, 
                STRUCT(FALSE as can_evolve, 5 as evolves_from) as evolution
            ),
            STRUCT(
                7 as id, "Squirtle" as name, ["Water"] as types, 
                STRUCT(TRUE as can_evolve, NULL as evolves_from) as evolution
            ),
            STRUCT(
                8 as id, "Wartortle" as name, ["Water"] as types, 
                STRUCT(TRUE as can_evolve, 7 as evolves_from) as evolution
            ),
            STRUCT(
                9 as id, "Blastoise" as name, ["Water"] as types, 
                STRUCT(FALSE as can_evolve, 8 as evolves_from) as evolution
            )
        ])
    """
    return bq.sql(query)



from typing import List

from google.cloud.bigquery import SchemaField

from bigquery_frame import DataFrame, BigQueryBuilder
from bigquery_frame.auth import get_bq_client
from bigquery_frame.dataframe import is_struct, is_repeated, is_nullable, schema_to_simple_string


def flatten(df: DataFrame, struct_separator: str = "_") -> DataFrame:
    """Flattens all the struct columns of a DataFrame
    Nested fields names will be joined together using the specified separator

    Examples:

    >>> bq = BigQueryBuilder(get_bq_client())
    >>> df = bq.sql('''SELECT 1 as id, STRUCT(1 as a, STRUCT(1 as c, 1 as d) as b) as s''')
    >>> df.printSchema()
    root
     |-- id: INTEGER (NULLABLE)
     |-- s: RECORD (NULLABLE)
     |    |-- a: INTEGER (NULLABLE)
     |    |-- b: RECORD (NULLABLE)
     |    |    |-- c: INTEGER (NULLABLE)
     |    |    |-- d: INTEGER (NULLABLE)
    <BLANKLINE>
    >>> flatten(df).printSchema()
    root
     |-- id: INTEGER (NULLABLE)
     |-- s_a: INTEGER (NULLABLE)
     |-- s_b_c: INTEGER (NULLABLE)
     |-- s_b_d: INTEGER (NULLABLE)
    <BLANKLINE>
    >>> flatten(df, "__").printSchema()
    root
     |-- id: INTEGER (NULLABLE)
     |-- s__a: INTEGER (NULLABLE)
     |-- s__b__c: INTEGER (NULLABLE)
     |-- s__b__d: INTEGER (NULLABLE)
    <BLANKLINE>

    :param df: a DataFrame
    :param struct_separator: It might be useful to change the separator when some DataFrame's column names already contain dots
    :return: a flattened DataFrame
    """
    # The idea is to recursively write a "SELECT s.b.c as s_b_c" for each nested column.
    cols = []

    def expand_struct(struct: List[SchemaField], col_stack: List[str]):
        for field in struct:
            if is_struct(field) and not is_repeated(field):
                expand_struct(field.fields, col_stack + [field.name])
            else:
                column = '.'.join(col_stack + [field.name]) + " as " + struct_separator.join(col_stack + [field.name])
                cols.append(column)

    expand_struct(df.schema, col_stack=[])
    return df.select(cols)


def flatten_schema(schema: List[SchemaField],
                   explode: bool,
                   struct_separator: str = ".",
                   array_separator: str = "!") -> List[SchemaField]:
    """Transforms a BigQuery DataFrame schema into a new schema where all structs have been flattened.
    The field names are kept, with a '.' separator for struct fields.
    If `explode` option is set, arrays are exploded with a '!' separator.

    Example:
    >>> bq = BigQueryBuilder(get_bq_client())
    >>> df = bq.sql('SELECT 1 as id, STRUCT(1 as a, [STRUCT(2 as c, 3 as d)] as b, [4, 5] as e) as s')
    >>> schema_to_simple_string(df.schema)
    'id:INTEGER,s:STRUCT<a:INTEGER,b:ARRAY<STRUCT<c:INTEGER,d:INTEGER>>,e:ARRAY<INTEGER>>'
    >>> schema_to_simple_string(flatten_schema(df.schema, explode=True))
    'id:INTEGER,s.a:INTEGER,s.b!.c:INTEGER,s.b!.d:INTEGER,s.e!:INTEGER'
    >>> schema_to_simple_string(flatten_schema(df.schema, explode=False))
    'id:INTEGER,s.a:INTEGER,s.b:ARRAY<STRUCT<c:INTEGER,d:INTEGER>>,s.e:ARRAY<INTEGER>'

    :param schema: A BigQuery DataFrame's schema
    :param explode: If set, arrays are exploded and an `array_separator` is appended to their name.
    :param struct_separator: separator used to delimit structs
    :param array_separator: separator used to delimit arrays
    :return:
    """
    def flatten_schema_field(prefix: str, schema_field: SchemaField, nullable: bool) -> List[SchemaField]:
        if is_struct(schema_field) and is_repeated(schema_field) and explode:
            return flatten_struct_type(schema_field.fields,
                                       nullable or is_nullable(schema_field),
                                       prefix + array_separator + struct_separator)
        elif is_struct(schema_field) and not is_repeated(schema_field):
            return flatten_struct_type(schema_field.fields,
                                       nullable or is_nullable(schema_field),
                                       prefix + struct_separator)
        else:
            mode = "NULLABLE" if nullable or is_nullable(schema_field) else "REQUIRED"
            if is_repeated(schema_field):
                if explode:
                    prefix += array_separator
                else:
                    mode = "REPEATED"
            return [SchemaField(prefix, schema_field.field_type, mode, schema_field.description, schema_field.fields)]

    def flatten_struct_type(schema: List[SchemaField], nullable: bool = False, prefix: str = "") -> List[SchemaField]:
        res = []
        for schema_field in schema:
            res += flatten_schema_field(prefix + schema_field.name, schema_field, nullable)
        return res

    return flatten_struct_type(schema)



from typing import Optional

from google.cloud.bigquery import QueryJobConfig, ConnectionProperty
from google.cloud.bigquery.client import Client
from google.cloud.bigquery.table import RowIterator


class HasBigQueryClient:
    """Wrapper class for the BigQuery client

    This isolates all the logic of direct interaction with the BigQuery client,
    which makes the code's security easier to audit (although nothing can be really private in Python).
    """

    def __init__(self, client: Client, use_session: bool = True):
        """Wrapper class for the BigQuery client

        :param client: A :class:`google.cloud.bigquery.client.Client`
        :param use_session: If set to true, all queries will be executed in the same session.
                            This is necessary for reusing temporary tables across multiple queries
        """
        self.__use_session = use_session
        self.__client = client
        self.__session_id: Optional[str] = None

    def _execute_query(self, query: str) -> RowIterator:
        job_config = QueryJobConfig()

        if self.__use_session and self.__session_id is None:
            job_config.create_session = True

        if self.__use_session and self.__session_id is not None:
            job_config.connection_properties = [
                ConnectionProperty("session_id", self.__session_id),
            ]

        job = self.__client.query(query=query, job_config=job_config)

        if job.error_result:
            return job.result()

        if self.__use_session and self.__session_id is None:
            self.__session_id = job.session_info.session_id

        return job.result()

    def close(self):
        self.__client.close()




import os
from typing import Optional

import google
from google.cloud.bigquery import Client
from google.oauth2 import service_account

from bigquery_frame import conf


def _get_bq_client_from_credentials() -> Optional[Client]:
    gcp_credentials_path = os.getenv("GCP_CREDENTIALS_PATH") or conf.GCP_CREDENTIALS_PATH
    if gcp_credentials_path.endswith(".json"):
        credentials = service_account.Credentials.from_service_account_file(
            gcp_credentials_path, scopes=["https://www.googleapis.com/auth/cloud-platform"],
        )
        client = google.cloud.bigquery.Client(credentials=credentials, project=credentials.project_id)
        return client


def _get_bq_client_from_default() -> Client:
    gcp_project = os.getenv("GCP_PROJECT") or conf.GCP_PROJECT
    client = google.cloud.bigquery.Client(gcp_project)
    return client


def get_bq_client():
    client = _get_bq_client_from_credentials()
    if client is None:
        client = _get_bq_client_from_default()
    return client




import re
from typing import List, Tuple, Optional, Union, Set

from google.cloud.bigquery import SchemaField, Client, Row
from google.cloud.bigquery.table import RowIterator

from bigquery_frame.auth import get_bq_client
from bigquery_frame.column import Column
from bigquery_frame.has_bigquery_client import HasBigQueryClient
from bigquery_frame.printing import print_results


Column = Union[str, Column]


def indent(str, nb) -> str:
    return " " * nb + str.replace("\n", "\n" + " " * nb)


def strip_margin(text):
    s = re.sub('\n[ \t]*\|', '\n', text)
    if s.startswith("\n"):
        return s[1:]
    else:
        return s


def quote(str) -> str:
    return "`" + str + "`"


def cols_to_str(cols, indentation: Optional[int] = None) -> str:
    cols = [str(col) for col in cols]
    if indentation is not None:
        return indent(",\n".join(cols), indentation)
    else:
        return ", ".join(cols)


def is_repeated(schema_field: SchemaField):
    return schema_field.mode == "REPEATED"


def is_struct(schema_field: SchemaField):
    return schema_field.field_type == "RECORD"


def is_nullable(schema_field: SchemaField):
    return schema_field.mode == "NULLABLE"


def schema_to_simple_string(schema: List[SchemaField]):
    """Transforms a BigQuery DataFrame schema into a new schema where all structs have been flattened.
    The field names are kept, with a '.' separator for struct fields.
    If `explode` option is set, arrays are exploded with a '!' separator.

    Example:

    >>> bq = BigQueryBuilder(get_bq_client())
    >>> df = bq.sql('SELECT 1 as id, STRUCT(1 as a, [STRUCT(2 as c, 3 as d)] as b, [4, 5] as e) as s')
    >>> print(df.schema)
    [SchemaField('id', 'INTEGER', 'NULLABLE', None, (), None), SchemaField('s', 'RECORD', 'NULLABLE', None, (SchemaField('a', 'INTEGER', 'NULLABLE', None, (), None), SchemaField('b', 'RECORD', 'REPEATED', None, (SchemaField('c', 'INTEGER', 'NULLABLE', None, (), None), SchemaField('d', 'INTEGER', 'NULLABLE', None, (), None)), None), SchemaField('e', 'INTEGER', 'REPEATED', None, (), None)), None)]
    >>> schema_to_simple_string(df.schema)
    'id:INTEGER,s:STRUCT<a:INTEGER,b:ARRAY<STRUCT<c:INTEGER,d:INTEGER>>,e:ARRAY<INTEGER>>'

    :param schema:
    :return:
    """
    def schema_field_to_simple_string(schema_field: SchemaField):
        if is_struct(schema_field):
            if is_repeated(schema_field):
                return f"ARRAY<STRUCT<{schema_to_simple_string(schema_field.fields)}>>"
            else:
                return f"STRUCT<{schema_to_simple_string(schema_field.fields)}>"
        else:
            if is_repeated(schema_field):
                return f"ARRAY<{schema_field.field_type}>"
            else:
                return schema_field.field_type

    cols = [f"{field.name}:{schema_field_to_simple_string(field)}" for field in schema]
    cols_to_string = ",".join(cols)
    return f"{cols_to_string}"


def schema_to_tree_string(schema: List[SchemaField]) -> str:
    """Generates a string representing the schema in tree format"""

    def str_gen_schema_field(schema_field: SchemaField, prefix: str) -> List[str]:
        res = [f"{prefix}{schema_field.name}: {schema_field.field_type} ({schema_field.mode})"]
        if is_struct(schema_field):
            res += str_gen_schema(schema_field.fields, " |   " + prefix)
        return res

    def str_gen_schema(schema: List[SchemaField], prefix: str) -> List[str]:
        return [
            str
            for schema_field in schema
            for str in str_gen_schema_field(schema_field, prefix)
        ]

    res = ["root"] + str_gen_schema(schema, " |-- ")

    return "\n".join(res) + "\n"


def _dedup_key_value_list(l: List[Tuple[object, object]]):
    """Deduplicate a list of key, values by their keys.
    Unlike `list(set(l))`, this does preserve ordering.

    Example:
    >>> _dedup_key_value_list([('a', 1), ('b', 2), ('a', 3)])
    [('a', 3), ('b', 2)]

    :param l:
    :return:
    """
    return list({k: v for k, v in l}.items())


class BigQueryBuilder(HasBigQueryClient):
    DEFAULT_ALIAS_NAME = "_default_alias_{num}"
    DEFAULT_TEMP_TABLE_NAME = "_default_temp_table_{num}"

    def __init__(self, client: Client):
        super().__init__(client)
        self._alias_count = 0
        self._temp_table_count = 0
        self._views: List[Tuple[str, 'DataFrame']] = []
        self._temp_tables: Set[str] = set()

    def table(self, full_table_name: str) -> 'DataFrame':
        """Returns the specified table as a :class:`DataFrame`.
        """
        query = f"""SELECT * FROM {quote(full_table_name)}"""
        return DataFrame(query, alias=None, bigquery=self)

    def sql(self, sql_query) -> 'DataFrame':
        """Returns a :class:`DataFrame` representing the result of the given query.
        """
        return DataFrame(sql_query, None, self)

    def _registerDataFrameAsTempView(self, df: 'DataFrame', alias: str) -> None:
        self._check_alias(alias, [])
        self._views.append((alias, df))

    def _registerDataFrameAsTempTable(self, df: 'DataFrame', alias: Optional[str]=None) -> 'DataFrame':
        if alias is None:
            alias = self._get_temp_table_alias()
        query = f"CREATE OR REPLACE TEMP TABLE {quote(alias)} AS \n" + df.compile()
        self._execute_query(query)
        return self.table(alias)

    def _compile_views(self) -> List[str]:
        return [
            strip_margin(f"""{quote(alias)} AS (
            |{indent(df._compile_with_deps(), 2)}
            |)""")
            for alias, df in self._views
        ]

    def _get_alias(self) -> str:
        self._alias_count += 1
        return self.DEFAULT_ALIAS_NAME.format(num=self._alias_count)

    def _get_temp_table_alias(self) -> str:
        self._temp_table_count += 1
        return self.DEFAULT_TEMP_TABLE_NAME.format(num=self._temp_table_count)

    def _check_alias(self, new_alias, deps: List[Tuple[str, 'DataFrame']]) -> None:
        """Checks that the alias follows BigQuery constraints, such as:

        - BigQuery does not allow having two CTEs with the same name in a query.

        :param new_alias:
        :param deps:
        :return: None
        :raises: an Exception if something that does not comply with BigQuery's rules is found.
        """
        collisions = [alias for alias, df in self._views + deps if alias == new_alias]
        if len(collisions) > 0:
            raise Exception(f"Duplicate alias {new_alias}")


class DataFrame:

    def __init__(self, query: str, alias: Optional[str], bigquery: BigQueryBuilder, deps: Optional[List['DataFrame']] = None):
        self.query = query
        if deps is None:
            deps = []
        deps = [dep for df in deps for dep in df._deps] + [(df._alias, df) for df in deps]
        self._deps: List[Tuple[str, 'DataFrame']] = _dedup_key_value_list(deps)
        if alias is None:
            alias = bigquery._get_alias()
        else:
            bigquery._check_alias(alias, self._deps)
        self._alias = alias
        self.bigquery: BigQueryBuilder = bigquery
        self._schema = None

    def __repr__(self):
        return f"""DataFrame('{self.query}) as {self._alias}')"""

    def _apply_query(self, query: str, deps: Optional[List['DataFrame']] = None) -> 'DataFrame':
        if deps is None:
            deps = [self]
        return DataFrame(query, None, self.bigquery, deps=deps)

    def _compute_schema(self):
        df = self.limit(0)
        return df.bigquery._execute_query(df.compile()).schema

    def _compile_deps(self) -> List[str]:
        return [
            strip_margin(f"""{quote(alias)} AS (
            |{indent(cte.query, 2)}
            |)""")
            for (alias, cte) in self._deps
        ]

    def _compile_with_ctes(self, ctes: List[str]) -> str:
        if len(ctes) > 0:
            return "WITH " + "\n, ".join(ctes) + "\n" + self.query
        else:
            return self.query

    def _compile_with_deps(self) -> str:
        ctes = self._compile_deps()
        return self._compile_with_ctes(ctes)

    @property
    def schema(self) -> List[SchemaField]:
        """Returns the schema of this :class:`DataFrame` as a list of :class:`google.cloud.bigquery.SchemaField`."""
        if self._schema is None:
            self._schema = self._compute_schema()
        return self._schema

    def compile(self) -> str:
        """Returns the sql query that will be executed to materialize this :class:`DataFrame`"""
        ctes = self.bigquery._compile_views() + self._compile_deps()
        return self._compile_with_ctes(ctes)

    def alias(self, alias) -> 'DataFrame':
        """Returns a new :class:`DataFrame` with an alias set."""
        return DataFrame(self.query, alias, self.bigquery)

    def persist(self) -> 'DataFrame':
        """Persist the contents of the :class:`DataFrame` in a temporary table and returns a new DataFrame reading
        from that table.

        Limitations compared to Spark
        -----------------------------
        Unlike with Spark, this operation is an action and the temporary table is created immediately.
        Like Spark, however, the current DataFrame is not persisted, the new DataFrame returned by this method must be used.

        :return: a new :class:`DataFrame`
        """
        return self.bigquery._registerDataFrameAsTempTable(self)

    def createOrReplaceTempTable(self, alias: str) -> 'DataFrame':
        """Creates or replace a persisted temporary table.

        >>> bq = BigQueryBuilder(get_bq_client())
        >>> df = bq.sql("SELECT 1 as id")
        >>> df.createOrReplaceTempTable("temp_table")
        DataFrame('SELECT * FROM `temp_table`) as _default_alias_2')
        >>> bq.sql("SELECT * FROM temp_table").show()
        +----+
        | id |
        +----+
        |  1 |
        +----+

        :return: a new :class:`DataFrame`
        """
        return self.bigquery._registerDataFrameAsTempTable(self, alias)

    def createOrReplaceTempView(self, name: str) -> None:
        """Creates or replaces a local temporary view with this :class:`DataFrame`.

        Limitations compared to Spark
        -----------------------------
        - Currently, replacing an existing temp view with a new one will raise an error.
          In this project, temporary views are implemented as CTEs in the final compiled query.
          As such, BigQuery does not allow to define two CTEs with the same name.

        :param name: Name of the temporary view. It must contain only alphanumeric and lowercase characters, no dots.
        :return: Nothing
        """
        self.bigquery._registerDataFrameAsTempView(self, name)

    def select(self, *columns: Union[List[Column], Column]) -> 'DataFrame':
        """Projects a set of expressions and returns a new :class:`DataFrame`."""
        if isinstance(columns[0], list):
            if len(columns) == 1:
                columns = columns[0]
            else:
                raise TypeError(f"Wrong argument type: {type(columns)}")
        query = strip_margin(
            f"""SELECT
            |{cols_to_str(columns, 2)}
            |FROM {quote(self._alias)}""")
        return self._apply_query(query)

    def drop(self, *cols: str) -> 'DataFrame':
        """Returns a new :class:`DataFrame` that drops the specified column.
        This is a no-op if schema doesn't contain the given column name(s).
        """
        schema_cols = set(self.columns)
        cols = [col for col in cols if col in schema_cols]
        query = strip_margin(
            f"""SELECT 
            |  * EXCEPT ({cols_to_str(cols)})
            |FROM {quote(self._alias)}""")
        return self._apply_query(query)

    def union(self, other: 'DataFrame') -> 'DataFrame':
        """Return a new :class:`DataFrame` containing union of rows in this and another :class:`DataFrame`.

        This is equivalent to `UNION ALL` in SQL. To do a SQL-style `UNION DISTINCT`
        (that does deduplication of elements), use this function followed by :func:`distinct`.

        Also as standard in SQL, this function resolves columns by position (not by name).
        To get column resolution by name, use the function `unionByName` instead.

        :param other:
        :return: a new :class:`DataFrame`
        """
        query = f"""SELECT * FROM {quote(self._alias)} UNION ALL SELECT * FROM {quote(other._alias)}"""
        return self._apply_query(query, deps=[self, other])

    unionAll = union

    def unionByName(self, other: 'DataFrame', allowMissingColumns: bool = False) -> 'DataFrame':
        """Returns a new :class:`DataFrame` containing union of rows in this and another :class:`DataFrame`.

        This is different from both `UNION ALL` and `UNION DISTINCT` in SQL. To do a SQL-style set
        union (that does deduplication of elements), use this function followed by :func:`distinct`.

        .. versionadded:: 2.3.0

        Examples
        --------
        The difference between this function and :func:`union` is that this function
        resolves columns by name (not by position):

        >>> bq = BigQueryBuilder(get_bq_client())
        >>> df1 = bq.sql('''SELECT 1 as col0, 2 as col1, 3 as col2''')
        >>> df2 = bq.sql('''SELECT 4 as col1, 5 as col2, 6 as col0''')
        >>> df1.union(df2).show()
        +------+------+------+
        | col0 | col1 | col2 |
        +------+------+------+
        |    1 |    2 |    3 |
        |    4 |    5 |    6 |
        +------+------+------+
        >>> df1.unionByName(df2).show()
        +------+------+------+
        | col0 | col1 | col2 |
        +------+------+------+
        |    1 |    2 |    3 |
        |    6 |    4 |    5 |
        +------+------+------+

        When the parameter `allowMissingColumns` is ``True``, the set of column names
        in this and other :class:`DataFrame` can differ; missing columns will be filled with null.
        Further, the missing columns of this :class:`DataFrame` will be added at the end
        in the schema of the union result:

        >>> df1 = bq.sql('''SELECT 1 as col0, 2 as col1, 3 as col2''')
        >>> df2 = bq.sql('''SELECT 4 as col1, 5 as col2, 6 as col3''')
        >>> df1.unionByName(df2, allowMissingColumns=True).show()
        +------+------+------+------+
        | col0 | col1 | col2 | col3 |
        +------+------+------+------+
        |    1 |    2 |    3 | null |
        | null |    4 |    5 |    6 |
        +------+------+------+------+

        :param other: Another DataFrame
        :param allowMissingColumns: If False, the columns that are not in both DataFrames are dropped, if True,
          they are added to the other DataFrame as null values.
        :return: a new :class:`DataFrame`
        """
        self_cols = self.columns
        other_cols = other.columns
        self_cols_set = set(self.columns)
        other_cols_set = set(other.columns)

        self_only_cols = [col for col in self_cols if col not in other_cols_set]
        common_cols = [col for col in self_cols if col in other_cols_set]
        other_only_cols = [col for col in other_cols if col not in self_cols_set]

        if (len(self_only_cols) > 0 or len(other_only_cols) > 0) and not allowMissingColumns:
            raise ValueError(f"UnionByName: dataFrames must have the same columns, "
                             f"unless allowMissingColumns is set to True.\n"
                             f"Columns in first DataFrame: [{cols_to_str(self.columns)}]\n"
                             f"Columns in second DataFrame: [{cols_to_str(other.columns)}]")

        def optional_comma(l: list):
            return ',' if len(l) > 0 else ''

        query = strip_margin(f"""
        |SELECT 
        |  {cols_to_str(self_only_cols, 2)}{optional_comma(self_only_cols)}
        |  {cols_to_str(common_cols, 2)}{optional_comma(common_cols)}
        |  {cols_to_str([f"NULL as {col}" for col in other_only_cols], 2)}
        |FROM {quote(self._alias)} 
        |UNION ALL 
        |SELECT 
        |  {cols_to_str([f"NULL as {col}" for col in self_only_cols], 2)}{optional_comma(self_only_cols)}
        |  {cols_to_str(common_cols, 2)}{optional_comma(common_cols)}
        |  {cols_to_str(other_only_cols, 2)}
        |FROM {quote(other._alias)}
        |""")
        return self._apply_query(query, deps=[self, other])

    def limit(self, num: int) -> 'DataFrame':
        """Returns a new :class:`DataFrame` with a result count limited to the specified number of rows."""
        query = f"""SELECT * FROM {quote(self._alias)} LIMIT {num}"""
        return self._apply_query(query)

    def distinct(self) -> 'DataFrame':
        """Returns a new :class:`DataFrame` containing the distinct rows in this :class:`DataFrame`.

        Limitations compared to Spark
        -----------------------------
        In BigQuery, the DISTINCT statement does not work on complex types like STRUCT and ARRAY.
        """
        query = f"""SELECT DISTINCT * FROM {quote(self._alias)}"""
        return self._apply_query(query)

    def sort(self, *cols: str):
        """Returns a new :class:`DataFrame` sorted by the specified column(s).

        :param cols:
        :return:
        """
        query = strip_margin(f"""
        |SELECT * 
        |FROM {quote(self._alias)} 
        |ORDER BY {cols_to_str(cols)}""")
        return self._apply_query(query)

    orderBy = sort

    def filter(self, expr: str) -> 'DataFrame':
        """Filters rows using the given condition."""
        query = strip_margin(f"""
        |SELECT * 
        |FROM {quote(self._alias)} 
        |WHERE {expr}""")
        return self._apply_query(query)

    where = filter

    def withColumn(self, col_name: str, col_expr: str, replace: bool = False) -> 'DataFrame':
        """Returns a new :class:`DataFrame` by adding a column or replacing the existing column that has the same name.

        The column expression must be an expression over this :class:`DataFrame`; attempting to add a column from
        some other :class:`DataFrame` will raise an error.

        Limitations compared to Spark
        -----------------------------
        - Replace an existing column must be done explicitly by passing the argument `replace=True`.
        - Each time this function is called, a new CTE will be created. This may lead to reaching BigQuery's
          query size limit very quickly.

        TODO: This project is just a POC. Future versions may bring improvements to these features but this will require more on-the-fly schema inspections.

        :param col_name: Name of the new column
        :param col_expr: Expression defining the new column
        :param replace: Set to true when replacing an already existing column
        :return: a new :class:`DataFrame`
        """
        if replace:
            query = f"SELECT * REPLACE ({col_expr} AS {col_name}) FROM {quote(self._alias)}"
        else:
            query = f"SELECT *, {col_expr} AS {col_name} FROM {quote(self._alias)}"
        return self._apply_query(query)

    def count(self) -> int:
        """Returns the number of rows in this :class:`DataFrame`."""
        query = f"SELECT COUNT(1) FROM {quote(self._alias)}"
        return self._apply_query(query).collect()[0][0]

    def collect_iterator(self) -> RowIterator:
        """Returns all the records as :class:`RowIterator`."""
        return self.bigquery._execute_query(self.compile())

    def collect(self) -> List[Row]:
        """Returns all the records as list of :class:`Row`."""
        return list(self.bigquery._execute_query(self.compile()))

    def take(self, num):
        """Returns the first ``num`` rows as a :class:`list` of :class:`Row`."""
        return self.limit(num).collect()

    def show(self, n: int = 20, format_args=None) -> None:
        """Prints the first ``n`` rows to the console. This uses the awesome Python library called `tabulate
        <https://pythonrepo.com/repo/astanin-python-tabulate-python-generating-and-working-with-logs>`_.

        Formating options may be set using `format_args`.

        >>> bq = BigQueryBuilder(get_bq_client())
        >>> df = bq.sql('''SELECT 1 as id, STRUCT(1 as a, [STRUCT(1 as c)] as b) as s''')
        >>> df.show()
        +----+---------------------------+
        | id |                         s |
        +----+---------------------------+
        |  1 | {'a': 1, 'b': [{'c': 1}]} |
        +----+---------------------------+
        >>> df.show(format_args={"tablefmt": 'fancy_grid'})
        ╒══════╤═══════════════════════════╕
        │   id │ s                         │
        ╞══════╪═══════════════════════════╡
        │    1 │ {'a': 1, 'b': [{'c': 1}]} │
        ╘══════╧═══════════════════════════╛

        :param n: Number of rows to show.
        :param format_args: extra arguments that may be passed to the function tabulate.tabulate()
        :return: Nothing
        """
        res = self.limit(n).collect_iterator()
        print_results(res, format_args)

    def toPandas(self):
        """Returns the contents of this :class:`DataFrame` as Pandas ``pandas.DataFrame``.

        This method requires to have following extra dependencies installed
        - pandas
        - pyarrow

        >>> df = __get_test_df()
        >>> import tabulate
        >>> pdf = df.toPandas()
        >>> type(pdf)
        <class 'pandas.core.frame.DataFrame'>
        >>> pdf
           id       name
        0   1  Bulbasaur
        1   2    Ivysaur
        2   3   Venusaur

        """
        return self.collect_iterator().to_dataframe()

    def treeString(self):
        """Generates a string representing the schema in tree format"""
        return schema_to_tree_string(self.schema)

    def printSchema(self) -> None:
        """Prints out the schema in tree format.

        Examples:

        >>> bq = BigQueryBuilder(get_bq_client())
        >>> df = bq.sql('''SELECT 1 as id, STRUCT(1 as a, [STRUCT(1 as c)] as b) as s''')
        >>> df.printSchema()
        root
         |-- id: INTEGER (NULLABLE)
         |-- s: RECORD (NULLABLE)
         |    |-- a: INTEGER (NULLABLE)
         |    |-- b: RECORD (REPEATED)
         |    |    |-- c: INTEGER (NULLABLE)
        <BLANKLINE>

        """
        print(self.treeString())

    def print_query(self) -> None:
        """Prints out the SQL query generated to materialize this :class:`DataFrame`.

        Examples:

        >>> bq = BigQueryBuilder(get_bq_client())
        >>> from bigquery_frame import functions as f
        >>> df = bq.sql('''SELECT 1 as a''').select('a', f.col('a')+f.lit(1).alias('b')).withColumn('c', f.expr('a + b'))
        >>> df.print_query()
        WITH `_default_alias_1` AS (
          SELECT 1 as a
        )
        , `_default_alias_2` AS (
          SELECT
            a,
            a + 1
          FROM `_default_alias_1`
        )
        SELECT *, a + b AS c FROM `_default_alias_2`

        """
        print(self.compile())

    @property
    def columns(self) -> List[str]:
        """Returns all column names as a list."""
        return [field.name for field in self.schema]


def __get_test_df() -> DataFrame:
    bq = BigQueryBuilder(get_bq_client())
    query = """
        SELECT 
            *
        FROM UNNEST ([
            STRUCT(1 as id, "Bulbasaur" as name),
            STRUCT(2 as id, "Ivysaur" as name),
            STRUCT(3 as id, "Venusaur" as name)
        ])
    """
    return bq.sql(query)



from typing import List

from bigquery_frame import DataFrame
from bigquery_frame.auth import get_bq_client
from bigquery_frame.dataframe import strip_margin, cols_to_str, BigQueryBuilder, quote


def pivot(df: DataFrame,
          pivot_column: str,
          agg_fun: str,
          agg_col: str,
          pivoted_columns: List[str] = None,
          implem_version:int = 1) -> DataFrame:
    """Pivots a column of the current :class:`DataFrame` and performs the specified aggregation.
    There are two versions of pivot function: one that requires the caller to specify the list
    of distinct values to pivot on, and one that does not. The latter is more concise but less
    efficient, because we need to first compute the list of distinct values internally.

    Example:
    >>> df = __get_test_unpivoted_df()
    >>> df.show()
    +------+---------+---------+--------+
    | year | product | country | amount |
    +------+---------+---------+--------+
    | 2018 |  Orange |  Canada |   null |
    | 2018 |  Orange |   China |   4000 |
    | 2018 |  Orange |  Mexico |   null |
    | 2018 |   Beans |  Canada |   null |
    | 2018 |   Beans |   China |   1500 |
    | 2018 |   Beans |  Mexico |   2000 |
    | 2018 |  Banana |  Canada |   2000 |
    | 2018 |  Banana |   China |    400 |
    | 2018 |  Banana |  Mexico |   null |
    | 2018 | Carrots |  Canada |   2000 |
    | 2018 | Carrots |   China |   1200 |
    | 2018 | Carrots |  Mexico |   null |
    | 2019 |  Orange |  Canada |   5000 |
    | 2019 |  Orange |   China |   null |
    | 2019 |  Orange |  Mexico |   5000 |
    | 2019 |   Beans |  Canada |   null |
    | 2019 |   Beans |   China |   1500 |
    | 2019 |   Beans |  Mexico |   2000 |
    | 2019 |  Banana |  Canada |   null |
    | 2019 |  Banana |   China |   1400 |
    +------+---------+---------+--------+
    >>> pivot(df, pivot_column="country", agg_fun="sum", agg_col="amount").show()
    +------+---------+--------+-------+--------+
    | year | product | Canada | China | Mexico |
    +------+---------+--------+-------+--------+
    | 2018 |  Orange |   null |  4000 |   null |
    | 2018 |   Beans |   null |  1500 |   2000 |
    | 2018 |  Banana |   2000 |   400 |   null |
    | 2018 | Carrots |   2000 |  1200 |   null |
    | 2019 |  Orange |   5000 |  null |   5000 |
    | 2019 |   Beans |   null |  1500 |   2000 |
    | 2019 |  Banana |   null |  1400 |    400 |
    | 2019 | Carrots |   null |   200 |   null |
    +------+---------+--------+-------+--------+

    :param df: a DataFrame
    :param pivot_column: column to pivot
    :param agg_fun: aggregation function that will be applied
    :param agg_col: column that will be aggregated
    :param pivoted_columns: (Optional) list of distinct values in the pivot column. Execution will be faster if provided.
    :param implem_version: (Possible values [1, 2]) Version of the code to use.
        Version 2 uses the BigQuery's PIVOT statement, while version 1 doesn't.
    :return:
    """
    if implem_version == 1:
        return pivot_v1(df, pivot_column, agg_fun, agg_col, pivoted_columns)
    elif implem_version == 2:
        return pivot_v2(df, pivot_column, agg_fun, agg_col, pivoted_columns)
    else:
        raise Exception("Pivot, please use v1 or v2")


def pivot_v1(df: DataFrame, pivot_column: str, agg_fun: str, agg_col: str, pivoted_columns: List[str] = None) -> DataFrame:
    """This version uses a good old GROUP BY statement and should be compatible with most ANSI-SQL engines."""
    group_columns = [col for col in df.columns if col.lower() not in [agg_col.lower(), pivot_column.lower()]]
    distinct_query = f"""SELECT DISTINCT {pivot_column} FROM {quote(df._alias)}"""
    if pivoted_columns is None:
        pivoted_columns = [row.get(pivot_column) for row in df._apply_query(distinct_query).collect()]

    aggregates = [f"""{agg_fun}(IF({pivot_column} = '{col}', {agg_col}, null)) as {col}""" for col in pivoted_columns]

    group_query = strip_margin(f"""
        |SELECT
        |{cols_to_str(group_columns, 2)},
        |{cols_to_str(aggregates, 2)}
        |FROM {quote(df._alias)}
        |GROUP BY {cols_to_str(group_columns)}
        |""")
    return df._apply_query(group_query)


def pivot_v2(df: DataFrame, pivot_column: str, agg_fun: str, agg_col: str, pivoted_columns: List[str] = None) -> DataFrame:
    """This version uses BigQuery's
    `PIVOT operator <https://cloud.google.com/bigquery/docs/reference/standard-sql/query-syntax#pivot_operator>`_
    """
    distinct_query = f"""SELECT DISTINCT {pivot_column} FROM {quote(df._alias)}"""
    if pivoted_columns is None:
        pivoted_columns = [row.get(pivot_column) for row in df._apply_query(distinct_query).collect()]
    quoted_pivoted_columns = [f"'{col}'" for col in pivoted_columns]

    group_query = strip_margin(f"""
        |SELECT
        | *
        |FROM {quote(df._alias)}
        |PIVOT({agg_fun}({agg_col}) FOR {pivot_column} IN ({cols_to_str(quoted_pivoted_columns)}))
        |""")
    return df._apply_query(group_query)


def unpivot(df: DataFrame,
            pivot_columns: List[str],
            key_alias: str = "key",
            value_alias: str = "value",
            exclude_nulls: bool = False,
            implem_version:int = 1) -> DataFrame:
    """Unpivot the given DataFrame along the specified pivot columns.
    All columns that are not pivot columns should have the same type.

    Example:
    >>> df = __get_test_pivoted_df()
    >>> df.show()
    +------+---------+--------+-------+--------+
    | year | product | Canada | China | Mexico |
    +------+---------+--------+-------+--------+
    | 2018 |  Orange |   null |  4000 |   null |
    | 2018 |   Beans |   null |  1500 |   2000 |
    | 2018 |  Banana |   2000 |   400 |   null |
    | 2018 | Carrots |   2000 |  1200 |   null |
    | 2019 |  Orange |   5000 |  null |   5000 |
    | 2019 |   Beans |   null |  1500 |   2000 |
    | 2019 |  Banana |   null |  1400 |    400 |
    | 2019 | Carrots |   null |   200 |   null |
    +------+---------+--------+-------+--------+
    >>> unpivot(df, ['year', 'product'], key_alias='country', value_alias='amount').show()
    +------+---------+---------+--------+
    | year | product | country | amount |
    +------+---------+---------+--------+
    | 2018 |  Orange |  Canada |   null |
    | 2018 |  Orange |   China |   4000 |
    | 2018 |  Orange |  Mexico |   null |
    | 2018 |   Beans |  Canada |   null |
    | 2018 |   Beans |   China |   1500 |
    | 2018 |   Beans |  Mexico |   2000 |
    | 2018 |  Banana |  Canada |   2000 |
    | 2018 |  Banana |   China |    400 |
    | 2018 |  Banana |  Mexico |   null |
    | 2018 | Carrots |  Canada |   2000 |
    | 2018 | Carrots |   China |   1200 |
    | 2018 | Carrots |  Mexico |   null |
    | 2019 |  Orange |  Canada |   5000 |
    | 2019 |  Orange |   China |   null |
    | 2019 |  Orange |  Mexico |   5000 |
    | 2019 |   Beans |  Canada |   null |
    | 2019 |   Beans |   China |   1500 |
    | 2019 |   Beans |  Mexico |   2000 |
    | 2019 |  Banana |  Canada |   null |
    | 2019 |  Banana |   China |   1400 |
    +------+---------+---------+--------+

    :param df: a DataFrame
    :param pivot_columns: The list of columns names on which to perform the pivot
    :param key_alias: alias given to the 'key' column
    :param value_alias: alias given to the 'value' column
    :param exclude_nulls: Exclude rows with null values if true
    :param implem_version: (Possible values [1, 2]) Version of the code to use.
         Version 2 uses the BigQuery's UNPIVOT statement, while version 1 doesn't.
    :return:
    """
    if implem_version == 1:
        return unpivot_v1(df, pivot_columns, key_alias, value_alias, exclude_nulls)
    elif implem_version == 2:
        return unpivot_v2(df, pivot_columns, key_alias, value_alias, exclude_nulls)
    else:
        raise Exception("Pivot, please use v1 or v2")


def unpivot_v1(df: DataFrame,
               pivot_columns: List[str],
               key_alias: str = "key",
               value_alias: str = "value",
               exclude_nulls: bool = False) -> DataFrame:
    """This version uses a LEFT JOIN UNNEST statement."""
    pivoted_columns = [(field.name, field.field_type) for field in df.schema if field.name not in pivot_columns]

    cols, types = zip(*pivoted_columns)

    # Check that all columns have the same type.
    assert len(
        set(types)) == 1, "All pivoted columns should be of the same type:\n Pivoted columns are: %s" % pivoted_columns

    # Create and explode an array of (column_name, column_value) structs
    struct_cols = [f'STRUCT("{col}" as {key_alias}, {col} as {value_alias})' for col in cols]
    exclude_nulls_str = f"WHERE {value_alias} IS NOT NULL" if exclude_nulls else ""
    query = strip_margin(f"""
        |SELECT
        |{cols_to_str(pivot_columns, 2)},
        |  pivoted.*
        |FROM {quote(df._alias)}
        |LEFT JOIN UNNEST([
        |{cols_to_str(struct_cols, 2)}
        |]) as pivoted
        |{exclude_nulls_str}
        |""")
    return df._apply_query(query)


def unpivot_v2(df: DataFrame,
               pivot_columns: List[str],
               key_alias: str = "key",
               value_alias: str = "value",
               exclude_nulls: bool = False) -> DataFrame:
    """This version uses BigQuery's
    `UNPIVOT operator <https://cloud.google.com/bigquery/docs/reference/standard-sql/query-syntax#unpivot_operator>`_
    """
    pivoted_columns = [(field.name, field.field_type) for field in df.schema if field.name not in pivot_columns]

    cols, types = zip(*pivoted_columns)

    exclude_nulls_str = f"EXCLUDE NULLS" if exclude_nulls else "INCLUDE NULLS"
    query = strip_margin(f"""
        |SELECT
        |  *
        |FROM {quote(df._alias)}
        |UNPIVOT {exclude_nulls_str}({value_alias} FOR {key_alias} IN ({cols_to_str(cols)}))
        |""")
    return df._apply_query(query)


def __get_test_pivoted_df() -> DataFrame:
    bq = BigQueryBuilder(get_bq_client())
    query = """
        SELECT 
            *
        FROM UNNEST ([
            STRUCT(2018 as year,  "Orange" as product, null as Canada, 4000 as China,  null as Mexico),
            STRUCT(2018 as year,   "Beans" as product, null as Canada, 1500 as China,  2000 as Mexico),
            STRUCT(2018 as year,  "Banana" as product, 2000 as Canada,  400 as China,  null as Mexico),
            STRUCT(2018 as year, "Carrots" as product, 2000 as Canada, 1200 as China,  null as Mexico),
            STRUCT(2019 as year,  "Orange" as product, 5000 as Canada, null as China,  5000 as Mexico),
            STRUCT(2019 as year,   "Beans" as product, null as Canada, 1500 as China,  2000 as Mexico),
            STRUCT(2019 as year,  "Banana" as product, null as Canada, 1400 as China,   400 as Mexico),
            STRUCT(2019 as year, "Carrots" as product, null as Canada,  200 as China,  null as Mexico)
        ])
    """
    return bq.sql(query)


def __get_test_unpivoted_df():
    bq = BigQueryBuilder(get_bq_client())
    query = """
        SELECT 
            *
        FROM UNNEST ([
            STRUCT(2018 as year, "Orange"  as product, "Canada" as country, null as amount),
            STRUCT(2018 as year, "Orange"  as product,  "China" as country, 4000 as amount),
            STRUCT(2018 as year, "Orange"  as product, "Mexico" as country, null as amount),
            STRUCT(2018 as year,  "Beans"  as product, "Canada" as country, null as amount),
            STRUCT(2018 as year,  "Beans"  as product,  "China" as country, 1500 as amount),
            STRUCT(2018 as year,  "Beans"  as product, "Mexico" as country, 2000 as amount),
            STRUCT(2018 as year, "Banana"  as product, "Canada" as country, 2000 as amount),
            STRUCT(2018 as year, "Banana"  as product,  "China" as country, 400 as amount),
            STRUCT(2018 as year, "Banana"  as product, "Mexico" as country, null as amount),
            STRUCT(2018 as year, "Carrots" as product, "Canada" as country, 2000 as amount),
            STRUCT(2018 as year, "Carrots" as product,  "China" as country, 1200 as amount),
            STRUCT(2018 as year, "Carrots" as product, "Mexico" as country, null as amount),
            STRUCT(2019 as year, "Orange"  as product, "Canada" as country, 5000 as amount),
            STRUCT(2019 as year, "Orange"  as product,  "China" as country, null as amount),
            STRUCT(2019 as year, "Orange"  as product, "Mexico" as country, 5000 as amount),
            STRUCT(2019 as year,  "Beans"  as product, "Canada" as country, null as amount),
            STRUCT(2019 as year,  "Beans"  as product,  "China" as country, 1500 as amount),
            STRUCT(2019 as year,  "Beans"  as product, "Mexico" as country, 2000 as amount),
            STRUCT(2019 as year, "Banana"  as product, "Canada" as country, null as amount),
            STRUCT(2019 as year, "Banana"  as product,  "China" as country, 1400 as amount),
            STRUCT(2019 as year, "Banana"  as product, "Mexico" as country, 400 as amount),
            STRUCT(2019 as year, "Carrots" as product, "Canada" as country, null as amount),
            STRUCT(2019 as year, "Carrots" as product,  "China" as country, 200 as amount),
            STRUCT(2019 as year, "Carrots" as product, "Mexico" as country, null as amount)
        ])
    """
    return bq.sql(query)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['bigquery_frame', 'bigquery_frame.transformations_impl']

package_data = \
{'': ['*']}

install_requires = \
['google-cloud-bigquery==2.31.0', 'tabulate==0.8.9', 'tqdm==4.62.3']

setup_kwargs = {
    'name': 'bigquery-frame',
    'version': '0.2.7',
    'description': 'A DataFrame API for Google BigQuery',
    'long_description': '# Bigquery-frame\n\n## What is it ?\n\nThis project is a POC that aims to showcase the wonders\nthat could be done if BigQuery provided a DataFrame API in \nPython similar to the one already available with PySpark\nor Snowpark (for which the Python API will come out soon).\n\nI tried to reproduce the most commonly used methods of the Spark DataFrame object. \nI aimed at making something as close as possible as PySpark, and tried to keep exactly\nthe same naming and docstrings as PySpark\'s DataFrames.\n \n\nFor instance, this is a working example of PySpark code :\n```python\nfrom pyspark.sql import SparkSession\nfrom pyspark.sql import functions as f\n\nspark = SparkSession.builder.master("local[1]").getOrCreate()\n\ndf = spark.sql("""\n    SELECT 1 as id, "Bulbasaur" as name, ARRAY("Grass", "Poison") as types, NULL as other_col\n    UNION ALL\n    SELECT 2 as id, "Ivysaur" as name, ARRAY("Grass", "Poison") as types, NULL as other_col\n""")\n\ndf.select("id", "name", "types").createOrReplaceTempView("pokedex")\n\ndf2 = spark.sql("""SELECT * FROM pokedex""")\\\n    .withColumn("nb_types", f.expr("SIZE(types)"))\\\n    .withColumn("name", f.expr("LOWER(name)"))\n\ndf2.show()\n# +---+---------+---------------+--------+\n# | id|     name|          types|nb_types|\n# +---+---------+---------------+--------+\n# |  1|bulbasaur|[Grass, Poison]|       2|\n# |  2|  ivysaur|[Grass, Poison]|       2|\n# +---+---------+---------------+--------+\n```\n\nAnd this is an equivalent working example using bigquery_frame, that runs on Google Big Query! \n```python\nfrom bigquery_frame import BigQueryBuilder\nfrom bigquery_frame.auth import get_bq_client\n\nbigquery = BigQueryBuilder(get_bq_client())\n\ndf = bigquery.sql("""\n    SELECT 1 as id, "Bulbasaur" as name, ["Grass", "Poison"] as types, NULL as other_col\n    UNION ALL\n    SELECT 2 as id, "Ivysaur" as name, ["Grass", "Poison"] as types, NULL as other_col\n""")\n\ndf.select("id", "name", "types").createOrReplaceTempView("pokedex")\n\ndf2 = bigquery.sql("""SELECT * FROM pokedex""")\\\n    .withColumn("nb_types", "ARRAY_LENGTH(types)")\\\n    .withColumn("name", "LOWER(name)", replace=True)\n\ndf2.show()\n# +----+-----------+---------------------+----------+\n# | id |   name    |        types        | nb_types |\n# +----+-----------+---------------------+----------+\n# | 1  | bulbasaur | [\'Grass\', \'Poison\'] |    2     |\n# | 2  |  ivysaur  | [\'Grass\', \'Poison\'] |    2     |\n# +----+-----------+---------------------+----------+\n```\n\n## What\'s so cool about DataFrames ?\n\nI believe that DataFrames are super cool to organise SQL code as it allows us to \nseveral things that are much harder, or even impossible, in pure-SQL:\n\n- on-the-fly introspection\n- chaining operations\n- generic transformations\n- higher level abstraction\n\nBut that deserves a blog article (coming soon).\n\n## I want to try this POC, how do I use it ?\n\nJust clone this repository, open PyCharm, and follow the\ninstructions in the [AUTH.md](/AUTH.md) documentation\nto set up your connection to BigQuery. Then, go fiddle\nwith the [demo](/examples/demo.py), or have a look at the [examples](/examples).\n\n\n## How does it work ?\n\nVery simply, by generating SQL queries that are sent to BigQuery.\nYou can get the query by calling the method `DataFrame.compile()`.\n\nFor instance, if we reuse the example from the beginning:\n```\nprint(df2.compile())\n```\n\nThis will print the following SQL query:\n```SQL\nWITH pokedex AS (\n  WITH _default_alias_1 AS (\n    \n        SELECT 1 as id, "Bulbasaur" as name, ["Grass", "Poison"] as types, NULL as other_col\n        UNION ALL\n        SELECT 2 as id, "Ivysaur" as name, ["Grass", "Poison"] as types, NULL as other_col\n    \n  )\n  SELECT \n    id,\n    name,\n    types\n  FROM _default_alias_1\n)\n, _default_alias_3 AS (\n  SELECT * FROM pokedex\n)\n, _default_alias_4 AS (\n  SELECT \n    *,\n    ARRAY_LENGTH(types) AS nb_types\n  FROM _default_alias_3\n)\nSELECT \n  * REPLACE (\n    LOWER(name) AS name\n  )\nFROM _default_alias_4\n```\n\n## Facturation\n\nThe examples in this code only use generated data and don\'t ready any "real" table.\nThis means that you won\'t be charged a cent running them.\n\nAlso, even when reading "real" tables, any one-the-fly introspection (such as\ngetting a DataFrame\'s schema), will trigger a query on BigQuery but will read\n0 rows, and will thus be billed 0 cent.\n\n## Known limitations\n\nSince this is a POC, I took some shortcuts and did not try to optimize the query length.\nIn particular, this uses _**a lot**_ of CTEs, and any serious project trying to use it\nmight reach the maximum query length very quickly.\n\nHere is a list of other known limitations, please also see the \n[Further developments](#further-developments) section for a list of missing features.\n\n- `DataFrame.withColumn`: \n  - unlike in Spark, replacing an existing column is  \n    not done automatically, an extra argument `replace=True` must be passed.\n- `DataFrame.createOrReplaceTempView`: \n  - I kept the same name as Spark for consistency, \n    but with the current implementation does not allow to replace an existing view\n  - Also, keep in mind this will not create an actual view on BigQuery, \n    it just emulates Spark\'s behaviour by using a CTE.\n\n## Further developments\n\nFunctions not supported yet :\n\n- `DataFrame.join`\n- `DataFrame.groupBy`\n- `DataFrame.printSchema`\n\nAlso, it would be cool to expand this to other SQL engines than BigQuery \n(contributors are welcome ;-) ).\n\n\n## Why did I make this ?\n\nI hope that it will motivate the teams working on BigQuery (or Redshift, \nor Azure Synapse) to propose a real python DataFrame API on top of their \nmassively parallel SQL engines. But not something ugly like this POC,\nthat generates SQL strings, more something like Spark Catalyst, which directly\ngenerates logical plans out of the DataFrame API without passing through the \n"SQL string" step.\n\nAfter starting this POC, I realized Snowflake already understood this and \ndeveloped Snowpark, a Java/Scala (and soon Python) API to run complex workflows\non Snowflake, and [Snowpark\'s DataFrame API](https://docs.snowflake.com/en/developer-guide/snowpark/reference/scala/com/snowflake/snowpark/DataFrame.html)\nwhich was clearly borrowed from [Spark\'s DataFrame (= DataSet[Row]) API](https://spark.apache.org/docs/latest/api/scala/org/apache/spark/sql/Dataset.html)\n(we recognize several key method names: cache, createOrReplaceTempView, \nwhere/filter, collect, toLocalIterator). \n\nI believe such project could open the gate to hundreds of very cool applications.\nFor instance, did you know that, in its early versions at least, Dataiku Shaker \nwas just a GUI that chained transformations on Pandas DataFrame, and later \nSpark DataFrame ? \n\nAnother possible evolution would be to make a DataFrame API capable of speaking\nmultiple SQL dialects. By using it, projects that generate SQL for multiple \nplatforms, like [Malloy](https://github.com/looker-open-source/malloy), could\nall use the same DataFrame abstraction. Adding support for a new SQL platform\nwould immediately allow all the project based on it to support this new platform.\n\n**I would be very interested if someone could make a similar POC with, \nRedShift, Postgres, Azure Synapse, or any other SQL engines \n(aside from Spark-SQL and Snowpark, of course :-p).**\n',
    'author': 'FurcyPin',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/FurcyPin/bigquery-frame',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.1,<3.11',
}


setup(**setup_kwargs)



from typing import Optional


def _bin_op(op: str):
    def fun(self, other: 'Column'):
        return Column(f"{self.expr} {op} {other.expr}")

    return fun


class Column:

    def __init__(self, expr: str, alias: Optional[str] = None):
        self.expr = expr
        self._alias = alias

    def __str__(self):
        if self._alias is None:
            return self.expr
        else:
            return f"{self.expr} as {self._alias}"

    def __repr__(self):
        return f"Column('{self.expr}')"

    __sub__ = _bin_op("-")
    __add__ = _bin_op("+")
    __mul__ = _bin_op("*")
    __div__ = _bin_op("/")

    def alias(self, alias: str) -> "Column":
        return Column(self.expr, alias)

    def asType(self, col_type: str) -> "Column":
        return Column(expr=f"CAST({self.expr} as {col_type})", alias=self._alias)
